@@ -2,11 +2,13 @@
 
 import os
 import json
+from pathlib import Path
 
 import docutils
 from docutils.parsers.rst import Directive, directives
-from docutils.nodes import math_block, image
+from docutils.nodes import math_block, image, literal
 from sphinx.util import parselinenos
+from sphinx.util.docutils import ReferenceRole
 from sphinx.addnodes import download_reference
 from sphinx.transforms import SphinxTransform
 from sphinx.environment.collectors.asset import ImageCollector
@@ -100,7 +102,7 @@ class JupyterCell(Directive):
                     location=location,
                 )
             try:
-                with open(filename) as f:
+                with Path(filename).open() as f:
                     content = [line.rstrip() for line in f.readlines()]
             except (IOError, OSError):
                 raise IOError("File {} not found or reading it failed".format(filename))
@@ -134,8 +136,6 @@ class JupyterCell(Directive):
             hide_code=("hide-code" in self.options),
             hide_output=("hide-output" in self.options),
             code_below=("code-below" in self.options),
-            linenos=("linenos" in self.options),
-            linenostart=(self.options.get("lineno-start")),
             emphasize_lines=hl_lines,
             raises=self.options.get("raises"),
             stderr=("stderr" in self.options),
@@ -144,7 +144,11 @@ class JupyterCell(Directive):
 
         # Add the input section of the cell, we'll add output at execution time
         cell_input = CellInputNode(classes=["cell_input"])
-        cell_input += docutils.nodes.literal_block(text="\n".join(content))
+        cell_input += docutils.nodes.literal_block(
+            text="\n".join(content),
+            linenos=("linenos" in self.options),
+            linenostart=(self.options.get("lineno-start")),
+        )
         cell_node += cell_input
         return [cell_node]
 
@@ -227,7 +231,8 @@ class JupyterWidgetStateNode(docutils.nodes.Element):
         )
 
 
-def cell_output_to_nodes(outputs, data_priority, write_stderr, dir, thebe_config):
+def cell_output_to_nodes(outputs, data_priority, write_stderr, out_dir,
+                         thebe_config, inline=False):
     """Convert a jupyter cell with outputs and filenames to doctree nodes.
 
     Parameters
@@ -237,17 +242,27 @@ def cell_output_to_nodes(outputs, data_priority, write_stderr, dir, thebe_config
         Which media types to prioritize.
     write_stderr : bool
         If True include stderr in cell output
-    dir : string
+    out_dir : string
         Sphinx "absolute path" to the output folder, so it is a relative path
         to the source folder prefixed with ``/``.
     thebe_config: dict
         Thebelab configuration object or None
+    inline: False
+        Whether the nodes will be placed in-line with the text.
 
     Returns
     -------
     to_add : list of docutils nodes
         Each output, converted into a docutils node.
     """
+    # If we're in `inline` mode, ensure that we don't add block-level nodes
+    if inline:
+        literal_node = docutils.nodes.literal
+        math_node = docutils.nodes.math
+    else:
+        literal_node = docutils.nodes.literal_block
+        math_node = math_block
+
     to_add = []
     for output in outputs:
         output_type = output["output_type"]
@@ -265,19 +280,22 @@ def cell_output_to_nodes(outputs, data_priority, write_stderr, dir, thebe_config
                     # Not setting "rawsource" disables Pygment hightlighting, which
                     # would otherwise add a <div class="highlight">.
 
-                    container = docutils.nodes.container(classes=["stderr"])
-                    container.append(
-                        docutils.nodes.literal_block(
-                            text=output["text"],
-                            rawsource="",  # disables Pygment highlighting
-                            language="none",
-                            classes=["stderr"],
-                        )
+                    literal = literal_node(
+                        text=output["text"],
+                        rawsource="",  # disables Pygment highlighting
+                        language="none",
+                        classes=["stderr"],
                     )
-                    to_add.append(container)
+                    if inline:
+                        # In this case, we don't wrap the text in containers
+                        to_add.append(literal)
+                    else:
+                        container = docutils.nodes.container(classes=["stderr"])
+                        container.append(literal)
+                        to_add.append(container)
             else:
                 to_add.append(
-                    docutils.nodes.literal_block(
+                    literal_node(
                         text=output["text"],
                         rawsource=output["text"],
                         language="none",
@@ -288,7 +306,7 @@ def cell_output_to_nodes(outputs, data_priority, write_stderr, dir, thebe_config
             traceback = "\n".join(output["traceback"])
             text = nbconvert.filters.strip_ansi(traceback)
             to_add.append(
-                docutils.nodes.literal_block(
+                literal_node(
                     text=text,
                     rawsource=text,
                     language="ipythontb",
@@ -303,19 +321,17 @@ def cell_output_to_nodes(outputs, data_priority, write_stderr, dir, thebe_config
                 continue
             data = output["data"][mime_type]
             if mime_type.startswith("image"):
+                file_path = Path(output.metadata["filenames"][mime_type])
+                out_dir = Path(out_dir)
                 # Sphinx treats absolute paths as being rooted at the source
                 # directory, so make a relative path, which Sphinx treats
                 # as being relative to the current working directory.
-                filename = os.path.basename(output.metadata["filenames"][mime_type])
+                filename = file_path.name
 
-                # checks if file dir path is inside a subdir of dir
-                filedir = os.path.dirname(output.metadata["filenames"][mime_type])
-                subpaths = filedir.split(dir)
-                if subpaths and len(subpaths) > 1:
-                    subpath = subpaths[1]
-                    dir += subpath
+                if out_dir in file_path.parents:
+                    out_dir = file_path.parent
 
-                uri = os.path.join(dir, filename)
+                uri = (out_dir / filename).as_posix()
                 to_add.append(docutils.nodes.image(uri=uri))
             elif mime_type == "text/html":
                 to_add.append(
@@ -325,7 +341,7 @@ def cell_output_to_nodes(outputs, data_priority, write_stderr, dir, thebe_config
                 )
             elif mime_type == "text/latex":
                 to_add.append(
-                    math_block(
+                    math_node(
                         text=strip_latex_delimiters(data),
                         nowrap=False,
                         number=None,
@@ -334,7 +350,7 @@ def cell_output_to_nodes(outputs, data_priority, write_stderr, dir, thebe_config
                 )
             elif mime_type == "text/plain":
                 to_add.append(
-                    docutils.nodes.literal_block(
+                    literal_node(
                         text=data,
                         rawsource=data,
                         language="none",
@@ -356,7 +372,7 @@ def cell_output_to_nodes(outputs, data_priority, write_stderr, dir, thebe_config
     return to_add
 
 
-def attach_outputs(output_nodes, node, thebe_config, cm_language):
+def attach_outputs(output_nodes, node, thebe_config):
     if not node.attributes["hide_code"]:  # only add css if code is displayed
         classes = node.attributes.get("classes", [])
         classes += ["jupyter_container"]
@@ -370,7 +386,7 @@ def attach_outputs(output_nodes, node, thebe_config, cm_language):
         thebe_source = ThebeSourceNode(
             hide_code=node.attributes["hide_code"],
             code_below=node.attributes["code_below"],
-            language=cm_language,
+            language=node.attributes["cm_language"],
         )
         thebe_source.children = [source]
         input_node.children = [thebe_source]
@@ -393,16 +409,19 @@ def attach_outputs(output_nodes, node, thebe_config, cm_language):
         node.children = node.children[::-1]
 
 
-def jupyter_download_role(name, rawtext, text, lineno, inliner):
-    _, filetype = name.split(":")
-    assert filetype in ("notebook", "script")
-    ext = ".ipynb" if filetype == "notebook" else ".py"
-    output_dir = sphinx_abs_dir(inliner.document.settings.env)
-    download_file = text + ext
-    node = download_reference(
-        download_file, download_file, reftarget=os.path.join(output_dir, download_file)
-    )
-    return [node], []
+class JupyterDownloadRole(ReferenceRole):
+    def run(self):
+        _, filetype = self.name.split(":")
+
+        assert filetype in ("notebook", "nb", "script")
+        ext = ".ipynb" if filetype in ("notebook", "nb") else ".py"
+        download_file = self.target + ext
+        reftarget = sphinx_abs_dir(self.env, download_file)
+        node = download_reference(self.rawtext, reftarget=reftarget)
+        self.set_source_info(node)
+        title = self.title if self.has_explicit_title else download_file
+        node += literal(self.rawtext, title, classes=["xref", "download"])
+        return [node], []
 
 
 def get_widgets(notebook):
@@ -434,7 +453,7 @@ class CellOutputsToNodes(SphinxTransform):
                 thebe_config,
             )
             # Remove the outputbundlenode and we'll attach the outputs next
-            attach_outputs(output_nodes, cell_node, thebe_config, cell_node.cm_language)
+            attach_outputs(output_nodes, cell_node, thebe_config)
 
         # Image collect extra nodes from cell outputs that we need to process
         for node in self.document.traverse(image):
